@@ -26,6 +26,7 @@
 
 #define _GNU_SOURCE
 
+#include <assert.h>
 #include <math.h>
 #include <stdlib.h>
 #include <stdint.h>
@@ -55,12 +56,12 @@ div_roundup(uint32_t n, size_t a)
 }
 
 struct wl_ring_buffer {
-	char data[4096];
-	uint32_t head, tail;
+	char *data;
+	size_t head, tail;
+	uint32_t size_bits;
+	uint32_t max_size_bits;  /* 0 for unlimited */
 };
 
-#define MASK(i) ((i) & 4095)
-
 #define MAX_FDS_OUT	28
 #define CLEN		(CMSG_LEN(MAX_FDS_OUT * sizeof(int32_t)))
 
@@ -71,23 +72,38 @@ struct wl_connection {
 	int want_flush;
 };
 
+static inline size_t
+size_pot(uint32_t size_bits)
+{
+	assert(size_bits < 8 * sizeof(size_t));
+
+	return ((size_t)1) << size_bits;
+}
+
+static size_t
+ring_buffer_capacity(const struct wl_ring_buffer *b) {
+	return size_pot(b->size_bits);
+}
+
+static size_t
+ring_buffer_mask(const struct wl_ring_buffer *b, size_t i) {
+	size_t m = ring_buffer_capacity(b) - 1;
+	return i & m;
+}
+
 static int
 ring_buffer_put(struct wl_ring_buffer *b, const void *data, size_t count)
 {
-	uint32_t head, size;
+	size_t head, size;
 
-	if (count > sizeof(b->data)) {
-		wl_log("Data too big for buffer (%d > %d).\n",
-		       count, sizeof(b->data));
-		errno = E2BIG;
-		return -1;
-	}
+	if (count == 0)
+		return 0;
 
-	head = MASK(b->head);
-	if (head + count <= sizeof b->data) {
+	head = ring_buffer_mask(b, b->head);
+	if (head + count <= ring_buffer_capacity(b)) {
 		memcpy(b->data + head, data, count);
 	} else {
-		size = sizeof b->data - head;
+		size = ring_buffer_capacity(b) - head;
 		memcpy(b->data + head, data, size);
 		memcpy(b->data, (const char *) data + size, count - size);
 	}
@@ -100,21 +116,21 @@ ring_buffer_put(struct wl_ring_buffer *b, const void *data, size_t count)
 static void
 ring_buffer_put_iov(struct wl_ring_buffer *b, struct iovec *iov, int *count)
 {
-	uint32_t head, tail;
+	size_t head, tail;
 
-	head = MASK(b->head);
-	tail = MASK(b->tail);
+	head = ring_buffer_mask(b, b->head);
+	tail = ring_buffer_mask(b, b->tail);
 	if (head < tail) {
 		iov[0].iov_base = b->data + head;
 		iov[0].iov_len = tail - head;
 		*count = 1;
 	} else if (tail == 0) {
 		iov[0].iov_base = b->data + head;
-		iov[0].iov_len = sizeof b->data - head;
+		iov[0].iov_len = ring_buffer_capacity(b) - head;
 		*count = 1;
 	} else {
 		iov[0].iov_base = b->data + head;
-		iov[0].iov_len = sizeof b->data - head;
+		iov[0].iov_len = ring_buffer_capacity(b) - head;
 		iov[1].iov_base = b->data;
 		iov[1].iov_len = tail;
 		*count = 2;
@@ -124,21 +140,21 @@ ring_buffer_put_iov(struct wl_ring_buffer *b, struct iovec *iov, int *count)
 static void
 ring_buffer_get_iov(struct wl_ring_buffer *b, struct iovec *iov, int *count)
 {
-	uint32_t head, tail;
+	size_t head, tail;
 
-	head = MASK(b->head);
-	tail = MASK(b->tail);
+	head = ring_buffer_mask(b, b->head);
+	tail = ring_buffer_mask(b, b->tail);
 	if (tail < head) {
 		iov[0].iov_base = b->data + tail;
 		iov[0].iov_len = head - tail;
 		*count = 1;
 	} else if (head == 0) {
 		iov[0].iov_base = b->data + tail;
-		iov[0].iov_len = sizeof b->data - tail;
+		iov[0].iov_len = ring_buffer_capacity(b) - tail;
 		*count = 1;
 	} else {
 		iov[0].iov_base = b->data + tail;
-		iov[0].iov_len = sizeof b->data - tail;
+		iov[0].iov_len = ring_buffer_capacity(b) - tail;
 		iov[1].iov_base = b->data;
 		iov[1].iov_len = head;
 		*count = 2;
@@ -148,26 +164,158 @@ ring_buffer_get_iov(struct wl_ring_buffer *b, struct iovec *iov, int *count)
 static void
 ring_buffer_copy(struct wl_ring_buffer *b, void *data, size_t count)
 {
-	uint32_t tail, size;
+	size_t tail, size;
 
-	tail = MASK(b->tail);
-	if (tail + count <= sizeof b->data) {
+	if (count == 0)
+		return;
+
+	tail = ring_buffer_mask(b, b->tail);
+	if (tail + count <= ring_buffer_capacity(b)) {
 		memcpy(data, b->data + tail, count);
 	} else {
-		size = sizeof b->data - tail;
+		size = ring_buffer_capacity(b) - tail;
 		memcpy(data, b->data + tail, size);
 		memcpy((char *) data + size, b->data, count - size);
 	}
 }
 
-static uint32_t
+static size_t
 ring_buffer_size(struct wl_ring_buffer *b)
 {
 	return b->head - b->tail;
 }
 
+static char *
+ring_buffer_tail(const struct wl_ring_buffer *b)
+{
+	return b->data + ring_buffer_mask(b, b->tail);
+}
+
+static uint32_t
+get_max_size_bits_for_size(size_t buffer_size)
+{
+	uint32_t max_size_bits = WL_BUFFER_DEFAULT_SIZE_POT;
+
+	/* buffer_size == 0 means unbound buffer size */
+	if (buffer_size == 0)
+		return 0;
+
+	while (max_size_bits < 8 * sizeof(size_t) && size_pot(max_size_bits) < buffer_size)
+		max_size_bits++;
+
+	return max_size_bits;
+}
+
+static int
+ring_buffer_allocate(struct wl_ring_buffer *b, size_t size_bits)
+{
+	char *new_data;
+
+	new_data = calloc(size_pot(size_bits), 1);
+	if (!new_data)
+		return -1;
+
+	ring_buffer_copy(b, new_data, ring_buffer_size(b));
+	free(b->data);
+	b->data = new_data;
+	b->size_bits = size_bits;
+	b->head = ring_buffer_size(b);
+	b->tail = 0;
+
+	return 0;
+}
+
+static size_t
+ring_buffer_get_bits_for_size(struct wl_ring_buffer *b, size_t net_size)
+{
+	size_t max_size_bits = get_max_size_bits_for_size(net_size);
+
+	if (max_size_bits < WL_BUFFER_DEFAULT_SIZE_POT)
+		max_size_bits = WL_BUFFER_DEFAULT_SIZE_POT;
+
+	if (b->max_size_bits > 0 && max_size_bits > b->max_size_bits)
+		max_size_bits = b->max_size_bits;
+
+	return max_size_bits;
+}
+
+static bool
+ring_buffer_is_max_size_reached(struct wl_ring_buffer *b)
+{
+	size_t net_size = ring_buffer_size(b) + 1;
+	size_t size_bits = ring_buffer_get_bits_for_size(b, net_size);
+
+	return net_size >= size_pot(size_bits);
+}
+
+static int
+ring_buffer_ensure_space(struct wl_ring_buffer *b, size_t count)
+{
+	size_t net_size = ring_buffer_size(b) + count;
+	size_t size_bits = ring_buffer_get_bits_for_size(b, net_size);
+
+	/* The 'size_bits' value represents the required size (in POT) to store
+	 * 'net_size', which depending whether the buffers are bounded or not
+	 * might not be sufficient (i.e. we might have reached the maximum size
+	 * allowed).
+	 */
+	if (net_size > size_pot(size_bits)) {
+		wl_log("Data too big for buffer (%d + %zd > %zd).\n",
+		       ring_buffer_size(b), count, size_pot(size_bits));
+		errno = E2BIG;
+		return -1;
+	}
+
+	/* The following test here is a short-cut to avoid reallocating a buffer
+	 * of the same size.
+	 */
+	if (size_bits == b->size_bits)
+		return 0;
+
+	/* Otherwise, we (re)allocate the buffer to match the required size */
+	return ring_buffer_allocate(b, size_bits);
+}
+
+static void
+ring_buffer_close_fds(struct wl_ring_buffer *buffer, int32_t count)
+{
+	int32_t i, *p;
+	size_t size, tail;
+
+	size = ring_buffer_capacity(buffer);
+	tail = ring_buffer_mask(buffer, buffer->tail);
+	p = (int32_t *) (buffer->data + tail);
+
+	for (i = 0; i < count; i++) {
+		if (p >= (int32_t *) (buffer->data + size))
+			p = (int32_t *) buffer->data;
+		close(*p++);
+	}
+}
+
+void
+wl_connection_set_max_buffer_size(struct wl_connection *connection,
+				  size_t max_buffer_size)
+{
+	uint32_t max_size_bits;
+
+	max_size_bits = get_max_size_bits_for_size(max_buffer_size);
+
+	connection->fds_in.max_size_bits = max_size_bits;
+	ring_buffer_ensure_space(&connection->fds_in, 0);
+
+	connection->fds_out.max_size_bits = max_size_bits;
+	ring_buffer_ensure_space(&connection->fds_out, 0);
+
+	connection->in.max_size_bits = max_size_bits;
+	ring_buffer_ensure_space(&connection->in, 0);
+
+	connection->out.max_size_bits = max_size_bits;
+	ring_buffer_ensure_space(&connection->out, 0);
+}
+
 struct wl_connection *
-wl_connection_create(int fd)
+wl_connection_create(int fd, size_t max_buffer_size)
 {
 	struct wl_connection *connection;
 
@@ -175,6 +323,8 @@ wl_connection_create(int fd)
 	if (connection == NULL)
 		return NULL;
 
+	wl_connection_set_max_buffer_size(connection, max_buffer_size);
+
 	connection->fd = fd;
 
 	return connection;
@@ -183,20 +333,20 @@ wl_connection_create(int fd)
 static void
 close_fds(struct wl_ring_buffer *buffer, int max)
 {
-	int32_t fds[sizeof(buffer->data) / sizeof(int32_t)], i, count;
 	size_t size;
+	int32_t count;
 
 	size = ring_buffer_size(buffer);
 	if (size == 0)
 		return;
 
-	ring_buffer_copy(buffer, fds, size);
-	count = size / sizeof fds[0];
+	count = size / sizeof(int32_t);
 	if (max > 0 && max < count)
 		count = max;
-	size = count * sizeof fds[0];
-	for (i = 0; i < count; i++)
-		close(fds[i]);
+
+	ring_buffer_close_fds(buffer, count);
+
+	size = count * sizeof(int32_t);
 	buffer->tail += size;
 }
 
@@ -212,7 +362,13 @@ wl_connection_destroy(struct wl_connection *connection)
 	int fd = connection->fd;
 
 	close_fds(&connection->fds_out, -1);
+	free(connection->fds_out.data);
+	free(connection->out.data);
+
 	close_fds(&connection->fds_in, -1);
+	free(connection->fds_in.data);
+	free(connection->in.data);
+
 	free(connection);
 
 	return fd;
@@ -256,7 +412,7 @@ static int
 decode_cmsg(struct wl_ring_buffer *buffer, struct msghdr *msg)
 {
 	struct cmsghdr *cmsg;
-	size_t size, max, i;
+	size_t size, i;
 	int overflow = 0;
 
 	for (cmsg = CMSG_FIRSTHDR(msg); cmsg != NULL;
@@ -266,8 +422,8 @@ decode_cmsg(struct wl_ring_buffer *buffer, struct msghdr *msg)
 			continue;
 
 		size = cmsg->cmsg_len - CMSG_LEN(0);
-		max = sizeof(buffer->data) - ring_buffer_size(buffer);
-		if (size > max || overflow) {
+
+		if (ring_buffer_ensure_space(buffer, size) < 0 || overflow) {
 			overflow = 1;
 			size /= sizeof(int32_t);
 			for (i = 0; i < size; i++)
@@ -293,17 +449,40 @@ wl_connection_flush(struct wl_connection *connection)
 	char cmsg[CLEN];
 	int len = 0, count;
 	size_t clen;
-	uint32_t tail;
+	size_t tail;
 
 	if (!connection->want_flush)
 		return 0;
 
 	tail = connection->out.tail;
-	while (connection->out.head - connection->out.tail > 0) {
-		ring_buffer_get_iov(&connection->out, iov, &count);
-
+	while (ring_buffer_size(&connection->out) > 0) {
 		build_cmsg(&connection->fds_out, cmsg, &clen);
 
+		if (clen >= CLEN) {
+			/* UNIX domain sockets allows to send file descriptors
+			 * using ancillary data.
+			 *
+			 * As per the UNIX domain sockets man page (man 7 unix),
+			 * "at least one byte of real data should be sent when
+			 * sending ancillary data".
+			 *
+			 * This is why we send only a single byte here, to ensure
+			 * all file descriptors are sent before the bytes are
+			 * cleared out.
+			 *
+			 * Otherwise This can fail to clear the file descriptors
+			 * first if individual messages are allowed to have 224
+			 * (8 bytes * MAX_FDS_OUT = 224) file descriptors .
+			 */
+			iov[0].iov_base = ring_buffer_tail(&connection->out);
+			iov[0].iov_len = 1;
+			count = 1;
+		} else {
+			ring_buffer_get_iov(&connection->out, iov, &count);
+		}
+
+		msg.msg_name = NULL;
+		msg.msg_namelen = 0;
 		msg.msg_iov = iov;
 		msg.msg_iovlen = count;
 		msg.msg_control = (clen > 0) ? cmsg : NULL;
@@ -341,49 +520,55 @@ wl_connection_read(struct wl_connection *connection)
 	char cmsg[CLEN];
 	int len, count, ret;
 
-	if (ring_buffer_size(&connection->in) >= sizeof(connection->in.data)) {
-		errno = EOVERFLOW;
-		return -1;
+	while (1) {
+		int data_size = ring_buffer_size(&connection->in);
+
+		/* Stop once we've read the max buffer size. */
+		if (ring_buffer_is_max_size_reached(&connection->in))
+			return data_size;
+
+		if (ring_buffer_ensure_space(&connection->in, 1) < 0)
+			return -1;
+
+		ring_buffer_put_iov(&connection->in, iov, &count);
+
+		msg.msg_name = NULL;
+		msg.msg_namelen = 0;
+		msg.msg_iov = iov;
+		msg.msg_iovlen = count;
+		msg.msg_control = cmsg;
+		msg.msg_controllen = sizeof cmsg;
+		msg.msg_flags = 0;
+
+		do {
+			len = wl_os_recvmsg_cloexec(connection->fd, &msg, MSG_DONTWAIT);
+		} while (len < 0 && errno == EINTR);
+
+		if (len == 0) {
+		    /* EOF, return previously read data first */
+		    return data_size;
+		}
+		if (len < 0) {
+		    if (errno == EAGAIN && data_size > 0) {
+			/* nothing new read, return previously read data */
+			return data_size;
+		    }
+		    return len;
+		}
+
+		ret = decode_cmsg(&connection->fds_in, &msg);
+		if (ret)
+			return -1;
+
+		connection->in.head += len;
 	}
-
-	ring_buffer_put_iov(&connection->in, iov, &count);
-
-	msg.msg_name = NULL;
-	msg.msg_namelen = 0;
-	msg.msg_iov = iov;
-	msg.msg_iovlen = count;
-	msg.msg_control = cmsg;
-	msg.msg_controllen = sizeof cmsg;
-	msg.msg_flags = 0;
-
-	do {
-		len = wl_os_recvmsg_cloexec(connection->fd, &msg, MSG_DONTWAIT);
-	} while (len < 0 && errno == EINTR);
-
-	if (len <= 0)
-		return len;
-
-	ret = decode_cmsg(&connection->fds_in, &msg);
-	if (ret)
-		return -1;
-
-	connection->in.head += len;
-
-	return wl_connection_pending_input(connection);
 }
 
 int
 wl_connection_write(struct wl_connection *connection,
 		    const void *data, size_t count)
 {
-	if (connection->out.head - connection->out.tail +
-	    count > ARRAY_LENGTH(connection->out.data)) {
-		connection->want_flush = 1;
-		if (wl_connection_flush(connection) < 0)
-			return -1;
-	}
-
-	if (ring_buffer_put(&connection->out, data, count) < 0)
+	if (wl_connection_queue(connection, data, count) < 0)
 		return -1;
 
 	connection->want_flush = 1;
@@ -395,13 +580,23 @@ int
 wl_connection_queue(struct wl_connection *connection,
 		    const void *data, size_t count)
 {
-	if (connection->out.head - connection->out.tail +
-	    count > ARRAY_LENGTH(connection->out.data)) {
+	/* We want to try to flush when the buffer reaches the default maximum
+	 * size even if the buffer has been previously expanded.
+	 *
+	 * Otherwise the larger buffer will cause us to flush less frequently,
+	 * which could increase lag.
+	 *
+	 * We'd like to flush often and get the buffer size back down if possible.
+	 */
+	if (ring_buffer_size(&connection->out) + count > WL_BUFFER_DEFAULT_MAX_SIZE) {
 		connection->want_flush = 1;
-		if (wl_connection_flush(connection) < 0)
+		if (wl_connection_flush(connection) < 0 && errno != EAGAIN)
 			return -1;
 	}
 
+	if (ring_buffer_ensure_space(&connection->out, count) < 0)
+		return -1;
+
 	return ring_buffer_put(&connection->out, data, count);
 }
 
@@ -411,7 +606,7 @@ wl_message_count_arrays(const struct wl_message *message)
 	int i, arrays;
 
 	for (i = 0, arrays = 0; message->signature[i]; i++) {
-		if (message->signature[i] == 'a')
+		if (message->signature[i] == WL_ARG_ARRAY)
 			arrays++;
 	}
 
@@ -427,12 +622,15 @@ wl_connection_get_fd(struct wl_connection *connection)
 static int
 wl_connection_put_fd(struct wl_connection *connection, int32_t fd)
 {
-	if (ring_buffer_size(&connection->fds_out) == MAX_FDS_OUT * sizeof fd) {
+	if (ring_buffer_size(&connection->fds_out) >= MAX_FDS_OUT * sizeof fd) {
 		connection->want_flush = 1;
-		if (wl_connection_flush(connection) < 0)
+		if (wl_connection_flush(connection) < 0 && errno != EAGAIN)
 			return -1;
 	}
 
+	if (ring_buffer_ensure_space(&connection->fds_out, sizeof fd) < 0)
+		return -1;
+
 	return ring_buffer_put(&connection->fds_out, &fd, sizeof fd);
 }
 
@@ -442,14 +640,14 @@ get_next_argument(const char *signature, struct argument_details *details)
 	details->nullable = 0;
 	for(; *signature; ++signature) {
 		switch(*signature) {
-		case 'i':
-		case 'u':
-		case 'f':
-		case 's':
-		case 'o':
-		case 'n':
-		case 'a':
-		case 'h':
+		case WL_ARG_INT:
+		case WL_ARG_UINT:
+		case WL_ARG_FIXED:
+		case WL_ARG_STRING:
+		case WL_ARG_OBJECT:
+		case WL_ARG_NEW_ID:
+		case WL_ARG_ARRAY:
+		case WL_ARG_FD:
 			details->type = *signature;
 			return signature + 1;
 		case '?':
@@ -466,14 +664,14 @@ arg_count_for_signature(const char *signature)
 	int count = 0;
 	for(; *signature; ++signature) {
 		switch(*signature) {
-		case 'i':
-		case 'u':
-		case 'f':
-		case 's':
-		case 'o':
-		case 'n':
-		case 'a':
-		case 'h':
+		case WL_ARG_INT:
+		case WL_ARG_UINT:
+		case WL_ARG_FIXED:
+		case WL_ARG_STRING:
+		case WL_ARG_OBJECT:
+		case WL_ARG_NEW_ID:
+		case WL_ARG_ARRAY:
+		case WL_ARG_FD:
 			++count;
 		}
 	}
@@ -506,32 +704,30 @@ wl_argument_from_va_list(const char *signature, union wl_argument *args,
 		sig_iter = get_next_argument(sig_iter, &arg);
 
 		switch(arg.type) {
-		case 'i':
+		case WL_ARG_INT:
 			args[i].i = va_arg(ap, int32_t);
 			break;
-		case 'u':
+		case WL_ARG_UINT:
 			args[i].u = va_arg(ap, uint32_t);
 			break;
-		case 'f':
+		case WL_ARG_FIXED:
 			args[i].f = va_arg(ap, wl_fixed_t);
 			break;
-		case 's':
+		case WL_ARG_STRING:
 			args[i].s = va_arg(ap, const char *);
 			break;
-		case 'o':
+		case WL_ARG_OBJECT:
 			args[i].o = va_arg(ap, struct wl_object *);
 			break;
-		case 'n':
+		case WL_ARG_NEW_ID:
 			args[i].o = va_arg(ap, struct wl_object *);
 			break;
-		case 'a':
+		case WL_ARG_ARRAY:
 			args[i].a = va_arg(ap, struct wl_array *);
 			break;
-		case 'h':
+		case WL_ARG_FD:
 			args[i].h = va_arg(ap, int32_t);
 			break;
-		case '\0':
-			return;
 		}
 	}
 }
@@ -545,7 +741,7 @@ wl_closure_clear_fds(struct wl_closure *closure)
 
 	for (i = 0; i < closure->count; i++) {
 		signature = get_next_argument(signature, &arg);
-		if (arg.type == 'h')
+		if (arg.type == WL_ARG_FD)
 			closure->args[i].h = -1;
 	}
 }
@@ -559,20 +755,27 @@ wl_closure_init(const struct wl_message *message, uint32_t size,
 
 	count = arg_count_for_signature(message->signature);
 	if (count > WL_CLOSURE_MAX_ARGS) {
-		wl_log("too many args (%d)\n", count);
+		wl_log("too many args (%d) for %s (signature %s)\n", count,
+		       message->name, message->signature);
 		errno = EINVAL;
 		return NULL;
 	}
 
+	int size_to_allocate;
+
 	if (size) {
 		*num_arrays = wl_message_count_arrays(message);
-		closure = zalloc(sizeof *closure + size +
-				 *num_arrays * sizeof(struct wl_array));
+		size_to_allocate = sizeof *closure + size +
+				   *num_arrays * sizeof(struct wl_array);
 	} else {
-		closure = zalloc(sizeof *closure);
+		size_to_allocate = sizeof *closure;
 	}
+	closure = zalloc(size_to_allocate);
 
 	if (!closure) {
+		wl_log("could not allocate closure of size (%d) for "
+		       "%s (signature %s)\n", size_to_allocate, message->name,
+		       message->signature);
 		errno = ENOMEM;
 		return NULL;
 	}
@@ -616,30 +819,30 @@ wl_closure_marshal(struct wl_object *sender, uint32_t opcode,
 		signature = get_next_argument(signature, &arg);
 
 		switch (arg.type) {
-		case 'f':
-		case 'u':
-		case 'i':
+		case WL_ARG_FIXED:
+		case WL_ARG_UINT:
+		case WL_ARG_INT:
 			break;
-		case 's':
+		case WL_ARG_STRING:
 			if (!arg.nullable && args[i].s == NULL)
 				goto err_null;
 			break;
-		case 'o':
+		case WL_ARG_OBJECT:
 			if (!arg.nullable && args[i].o == NULL)
 				goto err_null;
 			break;
-		case 'n':
+		case WL_ARG_NEW_ID:
 			object = args[i].o;
-			if (!arg.nullable && object == NULL)
+			if (object == NULL)
 				goto err_null;
 
 			closure->args[i].n = object ? object->id : 0;
 			break;
-		case 'a':
-			if (!arg.nullable && args[i].a == NULL)
+		case WL_ARG_ARRAY:
+			if (args[i].a == NULL)
 				goto err_null;
 			break;
-		case 'h':
+		case WL_ARG_FD:
 			fd = args[i].h;
 			dup_fd = wl_os_dupfd_cloexec(fd, 0);
 			if (dup_fd < 0) {
@@ -725,7 +928,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 	for (i = 0; i < count; i++) {
 		signature = get_next_argument(signature, &arg);
 
-		if (arg.type != 'h' && p + 1 > end) {
+		if (arg.type != WL_ARG_FD && p + 1 > end) {
 			wl_log("message too short, "
 			       "object (%d), message %s(%s)\n",
 			       closure->sender_id, message->name,
@@ -735,16 +938,16 @@ wl_connection_demarshal(struct wl_connection *connection,
 		}
 
 		switch (arg.type) {
-		case 'u':
+		case WL_ARG_UINT:
 			closure->args[i].u = *p++;
 			break;
-		case 'i':
+		case WL_ARG_INT:
 			closure->args[i].i = *p++;
 			break;
-		case 'f':
+		case WL_ARG_FIXED:
 			closure->args[i].f = *p++;
 			break;
-		case 's':
+		case WL_ARG_STRING:
 			length = *p++;
 
 			if (length == 0 && !arg.nullable) {
@@ -783,7 +986,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			closure->args[i].s = s;
 			p = next;
 			break;
-		case 'o':
+		case WL_ARG_OBJECT:
 			id = *p++;
 			closure->args[i].n = id;
 
@@ -795,11 +998,11 @@ wl_connection_demarshal(struct wl_connection *connection,
 				goto err;
 			}
 			break;
-		case 'n':
+		case WL_ARG_NEW_ID:
 			id = *p++;
 			closure->args[i].n = id;
 
-			if (id == 0 && !arg.nullable) {
+			if (id == 0) {
 				wl_log("NULL new ID received on non-nullable "
 				       "type, message %s(%s)\n", message->name,
 				       message->signature);
@@ -818,7 +1021,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			}
 
 			break;
-		case 'a':
+		case WL_ARG_ARRAY:
 			length = *p++;
 
 			length_in_u32 = div_roundup(length, sizeof *p);
@@ -839,7 +1042,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			closure->args[i].a = array_extra++;
 			p = next;
 			break;
-		case 'h':
+		case WL_ARG_FD:
 			if (connection->fds_in.tail == connection->fds_in.head) {
 				wl_log("file descriptor expected, "
 				       "object (%d), message %s(%s)\n",
@@ -902,35 +1105,35 @@ wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 	count = arg_count_for_signature(signature);
 	for (i = 0; i < count; i++) {
 		signature = get_next_argument(signature, &arg);
-		switch (arg.type) {
-		case 'o':
-			id = closure->args[i].n;
-			closure->args[i].o = NULL;
+		if (arg.type != WL_ARG_OBJECT)
+			continue;
 
-			object = wl_map_lookup(objects, id);
-			if (wl_object_is_zombie(objects, id)) {
-				/* references object we've already
-				 * destroyed client side */
-				object = NULL;
-			} else if (object == NULL && id != 0) {
-				wl_log("unknown object (%u), message %s(%s)\n",
-				       id, message->name, message->signature);
-				errno = EINVAL;
-				return -1;
-			}
+		id = closure->args[i].n;
+		closure->args[i].o = NULL;
 
-			if (object != NULL && message->types[i] != NULL &&
-			    !wl_interface_equal((object)->interface,
-						message->types[i])) {
-				wl_log("invalid object (%u), type (%s), "
-				       "message %s(%s)\n",
-				       id, (object)->interface->name,
-				       message->name, message->signature);
-				errno = EINVAL;
-				return -1;
-			}
-			closure->args[i].o = object;
+		object = wl_map_lookup(objects, id);
+		if (wl_object_is_zombie(objects, id)) {
+			/* references object we've already
+			 * destroyed client side */
+			object = NULL;
+		} else if (object == NULL && id != 0) {
+			wl_log("unknown object (%u), message %s(%s)\n",
+			       id, message->name, message->signature);
+			errno = EINVAL;
+			return -1;
 		}
+
+		if (object != NULL && message->types[i] != NULL &&
+		    !wl_interface_equal((object)->interface,
+					message->types[i])) {
+			wl_log("invalid object (%u), type (%s), "
+			       "message %s(%s)\n",
+			       id, (object)->interface->name,
+			       message->name, message->signature);
+			errno = EINVAL;
+			return -1;
+		}
+		closure->args[i].o = object;
 	}
 
 	return 0;
@@ -950,27 +1153,27 @@ convert_arguments_to_ffi(const char *signature, uint32_t flags,
 		sig_iter = get_next_argument(sig_iter, &arg);
 
 		switch(arg.type) {
-		case 'i':
+		case WL_ARG_INT:
 			ffi_types[i] = &ffi_type_sint32;
 			ffi_args[i] = &args[i].i;
 			break;
-		case 'u':
+		case WL_ARG_UINT:
 			ffi_types[i] = &ffi_type_uint32;
 			ffi_args[i] = &args[i].u;
 			break;
-		case 'f':
+		case WL_ARG_FIXED:
 			ffi_types[i] = &ffi_type_sint32;
 			ffi_args[i] = &args[i].f;
 			break;
-		case 's':
+		case WL_ARG_STRING:
 			ffi_types[i] = &ffi_type_pointer;
 			ffi_args[i] = &args[i].s;
 			break;
-		case 'o':
+		case WL_ARG_OBJECT:
 			ffi_types[i] = &ffi_type_pointer;
 			ffi_args[i] = &args[i].o;
 			break;
-		case 'n':
+		case WL_ARG_NEW_ID:
 			if (flags & WL_CLOSURE_INVOKE_CLIENT) {
 				ffi_types[i] = &ffi_type_pointer;
 				ffi_args[i] = &args[i].o;
@@ -979,11 +1182,11 @@ convert_arguments_to_ffi(const char *signature, uint32_t flags,
 				ffi_args[i] = &args[i].n;
 			}
 			break;
-		case 'a':
+		case WL_ARG_ARRAY:
 			ffi_types[i] = &ffi_type_pointer;
 			ffi_args[i] = &args[i].a;
 			break;
-		case 'h':
+		case WL_ARG_FD:
 			ffi_types[i] = &ffi_type_sint32;
 			ffi_args[i] = &args[i].h;
 			break;
@@ -1050,7 +1253,7 @@ copy_fds_to_connection(struct wl_closure *closure,
 	count = arg_count_for_signature(signature);
 	for (i = 0; i < count; i++) {
 		signature = get_next_argument(signature, &arg);
-		if (arg.type != 'h')
+		if (arg.type != WL_ARG_FD)
 			continue;
 
 		fd = closure->args[i].h;
@@ -1081,16 +1284,16 @@ buffer_size_for_closure(struct wl_closure *closure)
 		signature = get_next_argument(signature, &arg);
 
 		switch (arg.type) {
-		case 'h':
+		case WL_ARG_FD:
 			break;
-		case 'u':
-		case 'i':
-		case 'f':
-		case 'o':
-		case 'n':
+		case WL_ARG_UINT:
+		case WL_ARG_INT:
+		case WL_ARG_FIXED:
+		case WL_ARG_OBJECT:
+		case WL_ARG_NEW_ID:
 			buffer_size++;
 			break;
-		case 's':
+		case WL_ARG_STRING:
 			if (closure->args[i].s == NULL) {
 				buffer_size++;
 				break;
@@ -1099,7 +1302,7 @@ buffer_size_for_closure(struct wl_closure *closure)
 			size = strlen(closure->args[i].s) + 1;
 			buffer_size += 1 + div_roundup(size, sizeof(uint32_t));
 			break;
-		case 'a':
+		case WL_ARG_ARRAY:
 			if (closure->args[i].a == NULL) {
 				buffer_size++;
 				break;
@@ -1137,29 +1340,29 @@ serialize_closure(struct wl_closure *closure, uint32_t *buffer,
 	for (i = 0; i < count; i++) {
 		signature = get_next_argument(signature, &arg);
 
-		if (arg.type == 'h')
+		if (arg.type == WL_ARG_FD)
 			continue;
 
 		if (p + 1 > end)
 			goto overflow;
 
 		switch (arg.type) {
-		case 'u':
+		case WL_ARG_UINT:
 			*p++ = closure->args[i].u;
 			break;
-		case 'i':
+		case WL_ARG_INT:
 			*p++ = closure->args[i].i;
 			break;
-		case 'f':
+		case WL_ARG_FIXED:
 			*p++ = closure->args[i].f;
 			break;
-		case 'o':
+		case WL_ARG_OBJECT:
 			*p++ = closure->args[i].o ? closure->args[i].o->id : 0;
 			break;
-		case 'n':
+		case WL_ARG_NEW_ID:
 			*p++ = closure->args[i].n;
 			break;
-		case 's':
+		case WL_ARG_STRING:
 			if (closure->args[i].s == NULL) {
 				*p++ = 0;
 				break;
@@ -1174,7 +1377,7 @@ serialize_closure(struct wl_closure *closure, uint32_t *buffer,
 			memcpy(p, closure->args[i].s, size);
 			p += div_roundup(size, sizeof *p);
 			break;
-		case 'a':
+		case WL_ARG_ARRAY:
 			if (closure->args[i].a == NULL) {
 				*p++ = 0;
 				break;
@@ -1186,10 +1389,11 @@ serialize_closure(struct wl_closure *closure, uint32_t *buffer,
 			if (p + div_roundup(size, sizeof *p) > end)
 				goto overflow;
 
-			memcpy(p, closure->args[i].a->data, size);
+			if (size != 0)
+				memcpy(p, closure->args[i].a->data, size);
 			p += div_roundup(size, sizeof *p);
 			break;
-		default:
+		case WL_ARG_FD:
 			break;
 		}
 	}
@@ -1202,6 +1406,8 @@ serialize_closure(struct wl_closure *closure, uint32_t *buffer,
 	return size;
 
 overflow:
+	wl_log("serialize_closure overflow for %s (signature %s)\n",
+	       message->name, message->signature);
 	errno = ERANGE;
 	return -1;
 }
@@ -1219,8 +1425,13 @@ wl_closure_send(struct wl_closure *closure, struct wl_connection *connection)
 
 	buffer_size = buffer_size_for_closure(closure);
 	buffer = zalloc(buffer_size * sizeof buffer[0]);
-	if (buffer == NULL)
+	if (buffer == NULL) {
+		wl_log("wl_closure_send error: buffer allocation failure of "
+		       "size %d\n for %s (signature %s)",
+		       buffer_size * sizeof buffer[0], closure->message->name,
+		       closure->message->signature);
 		return -1;
+	}
 
 	size = serialize_closure(closure, buffer, buffer_size);
 	if (size < 0) {
@@ -1247,8 +1458,13 @@ wl_closure_queue(struct wl_closure *closure, struct wl_connection *connection)
 
 	buffer_size = buffer_size_for_closure(closure);
 	buffer = malloc(buffer_size * sizeof buffer[0]);
-	if (buffer == NULL)
+	if (buffer == NULL) {
+		wl_log("wl_closure_queue error: buffer allocation failure of "
+		       "size %d\n for %s (signature %s)",
+		       buffer_size * sizeof buffer[0], closure->message->name,
+		       closure->message->signature);
 		return -1;
+	}
 
 	size = serialize_closure(closure, buffer, buffer_size);
 	if (size < 0) {
@@ -1264,7 +1480,8 @@ wl_closure_queue(struct wl_closure *closure, struct wl_connection *connection)
 
 void
 wl_closure_print(struct wl_closure *closure, struct wl_object *target,
-		 int send, int discarded, uint32_t (*n_parse)(union wl_argument *arg))
+		 int send, int discarded, uint32_t (*n_parse)(union wl_argument *arg),
+		 const char *queue_name)
 {
 	int i;
 	struct argument_details arg;
@@ -1283,8 +1500,12 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target,
 	clock_gettime(CLOCK_REALTIME, &tp);
 	time = (tp.tv_sec * 1000000L) + (tp.tv_nsec / 1000);
 
-	fprintf(f, "[%7u.%03u] %s%s%s@%u.%s(",
-		time / 1000, time % 1000,
+	fprintf(f, "[%7u.%03u] ", time / 1000, time % 1000);
+
+	if (queue_name)
+		fprintf(f, "{%s} ", queue_name);
+
+	fprintf(f, "%s%s%s#%u.%s(",
 		discarded ? "discarded " : "",
 		send ? " -> " : "",
 		target->interface->name, target->id,
@@ -1296,13 +1517,13 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target,
 			fprintf(f, ", ");
 
 		switch (arg.type) {
-		case 'u':
+		case WL_ARG_UINT:
 			fprintf(f, "%u", closure->args[i].u);
 			break;
-		case 'i':
+		case WL_ARG_INT:
 			fprintf(f, "%d", closure->args[i].i);
 			break;
-		case 'f':
+		case WL_ARG_FIXED:
 			/* The magic number 390625 is 1e8 / 256 */
 			if (closure->args[i].f >= 0) {
 				fprintf(f, "%d.%08d",
@@ -1315,27 +1536,27 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target,
 					-390625 * (closure->args[i].f % 256));
 			}
 			break;
-		case 's':
+		case WL_ARG_STRING:
 			if (closure->args[i].s)
 				fprintf(f, "\"%s\"", closure->args[i].s);
 			else
 				fprintf(f, "nil");
 			break;
-		case 'o':
+		case WL_ARG_OBJECT:
 			if (closure->args[i].o)
-				fprintf(f, "%s@%u",
+				fprintf(f, "%s#%u",
 					closure->args[i].o->interface->name,
 					closure->args[i].o->id);
 			else
 				fprintf(f, "nil");
 			break;
-		case 'n':
+		case WL_ARG_NEW_ID:
 			if (n_parse)
 				nval = n_parse(&closure->args[i]);
 			else
 				nval = closure->args[i].n;
 
-			fprintf(f, "new id %s@",
+			fprintf(f, "new id %s#",
 				(closure->message->types[i]) ?
 				 closure->message->types[i]->name :
 				  "[unknown]");
@@ -1344,10 +1565,10 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target,
 			else
 				fprintf(f, "nil");
 			break;
-		case 'a':
+		case WL_ARG_ARRAY:
 			fprintf(f, "array[%zu]", closure->args[i].a->size);
 			break;
-		case 'h':
+		case WL_ARG_FD:
 			fprintf(f, "fd %d", closure->args[i].h);
 			break;
 		}
@@ -1370,7 +1591,7 @@ wl_closure_close_fds(struct wl_closure *closure)
 
 	for (i = 0; i < closure->count; i++) {
 		signature = get_next_argument(signature, &arg);
-		if (arg.type == 'h' && closure->args[i].h != -1)
+		if (arg.type == WL_ARG_FD && closure->args[i].h != -1)
 			close(closure->args[i].h);
 	}
 
