@@ -24,6 +24,7 @@
  * SOFTWARE.
  */
 
+#define _GNU_SOURCE
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -1005,9 +1006,16 @@ registry_handle_filtered(void *data, struct wl_registry *registry,
 	}
 }
 
+static void
+registry_handle_remove_filtered(void *data, struct wl_registry *registry,
+				uint32_t id)
+{
+	assert(false);
+}
+
 static const struct wl_registry_listener registry_listener_filtered = {
 	registry_handle_filtered,
-	NULL
+	registry_handle_remove_filtered,
 };
 
 static void
@@ -1043,6 +1051,58 @@ TEST(filtered_global_is_hidden)
 	display_destroy(d);
 }
 
+static void
+get_dynamic_globals(void *data)
+{
+	struct client *c = client_connect();
+	struct wl_registry *registry;
+
+	registry = wl_display_get_registry(c->wl_display);
+	wl_registry_add_listener(registry, &registry_listener_filtered, data);
+	wl_display_roundtrip(c->wl_display);
+
+	/* Wait for the server to create a new global */
+	assert(stop_display(c, 1) >= 0);
+
+	/* Check that we don't see it */
+	wl_display_roundtrip(c->wl_display);
+
+	/* Wait for the server to remove that global */
+	assert(stop_display(c, 1) >= 0);
+
+	/* Check that we don't get a global_remove event */
+	wl_display_roundtrip(c->wl_display);
+
+	wl_registry_destroy(registry);
+	client_disconnect_nocheck(c);
+}
+
+TEST(filtered_dynamic_global_is_hidden)
+{
+	struct display *d;
+	struct wl_global *g;
+
+	d = display_create();
+	wl_display_set_global_filter(d->wl_display, global_filter, NULL);
+
+	/* Create a client and let it enumerate the globals */
+	client_create_noarg(d, get_dynamic_globals);
+	display_run(d);
+
+	/* Dynamically create a new global */
+	g = wl_global_create(d->wl_display, &wl_data_offer_interface,
+			     1, d, bind_data_offer);
+
+	display_resume(d);
+
+	/* Dynamically remove the global */
+	wl_global_destroy(g);
+
+	display_resume(d);
+
+	display_destroy(d);
+}
+
 static void
 check_bind_error(struct client *c)
 {
@@ -1432,6 +1492,10 @@ send_overflow_client(void *data)
 	char tmp = '\0';
 	int sock, optval = 16384;
 
+	/* By default, client buffers are now unbounded, set a limit to cause
+	 * an overflow, otherwise the client buffers will grow indefinitely. */
+	wl_display_set_max_buffer_size(c->wl_display, 4096);
+
 	/* Limit the send buffer size for the display socket to guarantee
 	 * that the test will cause an overflow. */
 	sock = wl_display_get_fd(c->wl_display);
@@ -1445,6 +1509,7 @@ send_overflow_client(void *data)
 	 * within <=4096 iterations. */
 	for (i = 0; i < 1000000; i++) {
 		noop_request(c);
+		fprintf(stderr, "Send loop %i\n", i);
 		err = wl_display_get_error(c->wl_display);
 		if (err)
 			break;
@@ -1454,9 +1519,9 @@ send_overflow_client(void *data)
 	 * check verifies that the initial/final FD counts are the same */
 	assert(write(pipes[1], &tmp, sizeof(tmp)) == (ssize_t)sizeof(tmp));
 
-	/* Expect an error */
+	/* Expect an error - ring_buffer_ensure_space() returns E2BIG */
 	fprintf(stderr, "Send loop failed on try %d, err = %d, %s\n", i, err, strerror(err));
-	assert(err == EAGAIN);
+	assert(err == EAGAIN || err == E2BIG);
 
 	client_disconnect_nocheck(c);
 }
@@ -1629,3 +1694,24 @@ TEST(global_remove)
 
 	display_destroy(d);
 }
+
+static void
+terminate_display(void *arg)
+{
+	struct wl_display *wl_display = arg;
+	wl_display_terminate(wl_display);
+}
+
+TEST(no_source_terminate)
+{
+	struct display *d;
+	struct wl_event_loop *loop;
+
+	d = display_create();
+	loop = wl_display_get_event_loop(d->wl_display);
+
+	wl_event_loop_add_idle(loop, terminate_display, d->wl_display);
+
+	display_run(d);
+	display_destroy(d);
+}
